@@ -2,18 +2,21 @@
 Contains CS basic operations and pipelines
 '''
 import os, sys
+from tqdm import tqdm
 import numpy as np
 import pandas as pd
 import scipy
 import cv2
+import cvxpy
 import matplotlib
 import matplotlib.pyplot as plt
 from matplotlib import cm
 import pylab
 
+from sklearn.preprocessing import StandardScaler
 from sklearn.decomposition import PCA
 from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
-from sklearn.linear_model import LassoCV, OrthogonalMatchingPursuit, OrthogonalMatchingPursuitCV
+from sklearn.linear_model import Lasso, LassoCV, LogisticRegressionCV, OrthogonalMatchingPursuit
 from statsmodels.multivariate.manova import MANOVA
 
 if __package__:
@@ -37,7 +40,7 @@ def Sensing(x, k = 0.2, flavor = 'identity'):
     ----------
     k : (0,1]. sampling ratio.
     flavor : 
-        'identity' - use a scramble identity matrix. take its n*k rows as sensing matrix
+        'identity' / 'bernoulli' - use a scramble identity matrix. take its n*k rows as sensing matrix
         'gaussian' - sampling from N(1, 1/n)
 
     Return
@@ -67,7 +70,7 @@ def MeasurementMatrix (N, r, t = 'DCT'):
     '''
     
     kn = len(r)    
-    a = np.zeros(N)    
+    a = np.zeros((1, N))  
     t = t.upper()
         
     # Suppose the first number in the permutation is 42. Then choose the 42th
@@ -79,14 +82,14 @@ def MeasurementMatrix (N, r, t = 'DCT'):
     A = np.zeros ((kn, N)) #csr_matrix, dtype='float16') # reduce memory use. numpy uses float64 as default dtype.
 
     for i, j in enumerate(r):
-        a[j] = 1
+        a[0,j] = 1
         if t == 'IDM' or t == 0:
-            A[i] = a # .astype(np.float16)
+            A[i,:] = a # .astype(np.float16)
         elif t == 'DCT' or t == 1:
-            A[i] = cv2.dct(a).flatten() # .astype(np.float16)
+            A[i,:] = cv2.dct(a)# .flatten() # .astype(np.float16)
         elif t == 'DFT' or t == 2:
-            A[i] = cv2.dft(a).flatten() # .astype(np.float16)
-        a[j] = 0
+            A[i,:] = cv2.dft(a)# .flatten() # .astype(np.float16)
+        a[0,j] = 0
     
     ###### ALTERNATIVE IMPLEMENTATION ######
     # phi = np.zeros((k, N)) # Sampling matrix, kxN    
@@ -97,7 +100,7 @@ def MeasurementMatrix (N, r, t = 'DCT'):
     
     return A # .toarray()
 
-def Recovery (A, xs, t = 'DCT', PSI = None, solver = 'LASSO', fast_lasso = False, \
+def Recovery (A, xs, t = 'DCT', PSI = None, solver = 'LASSO', L1 = 0.005, \
     display = True, verbose = False):
     '''    
     Solve the optimization problem A@z = y
@@ -110,7 +113,8 @@ def Recovery (A, xs, t = 'DCT', PSI = None, solver = 'LASSO', fast_lasso = False
                       Now we use LassoCV. This param is no longer needed.
     t : IDM, DCT, DFT, etc.
     PSI : For adpative transforms, users need to pass in the PSI basis. For non-adaptive ones, leave as none.
-    solver : 'LASSO' or 'OMP'
+    solver : 'LASSO' or 'OMP' or 'BP' or 'VAE' (variational autoencoder, need a pretrained VAE model)
+    L1 : LASSO L1. specify a float or leave None for CV.
     '''
 
     t = t.upper()
@@ -122,18 +126,18 @@ def Recovery (A, xs, t = 'DCT', PSI = None, solver = 'LASSO', fast_lasso = False
             t == 'DFT' or \
             t == 'DWT' or \
             t == 'LDA':
-            alphas = [0.0001, 0.00001] # set empirical alpha values. LDA needs smaller alpha
+            alphas = [0.005, 0.0001, 0.00001] # set empirical alpha values. LDA needs smaller alpha
 
-        # 实测并未发现两种模式的运行时间差异
-        if fast_lasso:
-            lasso = LassoCV(alphas = alphas, selection = 'random', tol = 0.001, n_jobs=-1, verbose = verbose) # ‘random’ often leads to significantly faster convergence especially when tol is higher than 1e-4.
+        if isinstance(L1, float):
+            lasso = Lasso(alpha=L1, tol = 0.0005)    
         else:
-            lasso = LassoCV(alphas = alphas, selection = 'cyclic', n_jobs=-1, verbose = verbose)
-
+            lasso = LassoCV(alphas = alphas, selection = 'random', tol = 0.001, n_jobs=-1, verbose = verbose) # ‘random’ often leads to significantly faster convergence especially when tol is higher than 1e-4.
+            # 实测并未发现 selction = cyclic / random 两种模式的运行时间差异
+        
         lasso.fit(A, xs)
         z = lasso.coef_
 
-    else: # 'OMP'
+    elif solver == 'OMP':
 
         # plot the noise-free reconstruction
         omp = OrthogonalMatchingPursuit() # n_nonzero_coefs default 10%.
@@ -141,6 +145,16 @@ def Recovery (A, xs, t = 'DCT', PSI = None, solver = 'LASSO', fast_lasso = False
         z = omp.coef_
         # print('OMP score:', omp.score(A, xs)) # Return the coefficient of determination R^2 of the prediction.
     
+    elif solver == 'BP':
+        
+        n = A.shape[1]
+        z = cvxpy.Variable(n)
+        obj = cvxpy.Minimize(cvxpy.norm(z, 1))
+        const = [A @ z == xs]
+        prob = cvxpy.Problem(obj,const)
+        _ = prob.solve()       
+        z = z.value
+
     if t == PSI_NAMES[0] or t == PSI_LONGNAMES[0]:
         z = np.linalg.pinv(A) @ xs
         xr = z
@@ -165,7 +179,7 @@ def Recovery (A, xs, t = 'DCT', PSI = None, solver = 'LASSO', fast_lasso = False
         plt.figure(figsize = (10,3))
         # print ('non-zero coef: ', np.count_nonzero(z))
         # print ('sparsity: ', 1 - np.count_nonzero(z) / len(z) )
-        biggest_lasso_fs = (np.argsort(np.abs(z))[-200:-1])[::-1] # take last N item indices and reverse (ord desc)
+        _ = (np.argsort(np.abs(z))[-200:-1])[::-1] # take last N item indices and reverse (ord desc)
         plt.plot(np.abs(z))
         plt.title('abs (z)')
         plt.show()
@@ -178,7 +192,7 @@ def Recovery (A, xs, t = 'DCT', PSI = None, solver = 'LASSO', fast_lasso = False
     return z, xr
 
 def Sensing_n_Recovery(x, k = 0.2, t = 'DCT', solver = 'LASSO', \
-    fast_lasso = False, display = True):
+    L1 = 0.005, display = True):
     
     '''
     Provide a complete pipeline for signal sensing and recovery with a speicfic PSI. 
@@ -189,7 +203,7 @@ def Sensing_n_Recovery(x, k = 0.2, t = 'DCT', solver = 'LASSO', \
         In normal cases (k<1), Az = xs has more unknowns than equations. For the extreme case of k = 1, CS sampling is degraded to a random shuffling. The linear system 'Az = xs' will have equal unknowns and equations, and there is only one solution. The reconstructed xr will be identical to the original x.
     '''
 
-    VECTORIZATION = False    
+    VECTORIZATION = False  
     if VECTORIZATION:
         PHI, OMEGA = GetSensingMatrix(len(x), k) 
         xs = PHI @ x
@@ -204,28 +218,10 @@ def Sensing_n_Recovery(x, k = 0.2, t = 'DCT', solver = 'LASSO', \
         xs, r = Sensing(x, k)
         A = MeasurementMatrix (len(x), r, t = t)
 
-    z,xr = Recovery (A, xs, t = t, solver = solver, fast_lasso=fast_lasso, display = False)
+    z,xr = Recovery (A, xs, t = t, solver = solver, L1=L1, display = False)
 
-    if display: 
-
-        matplotlib.rcParams.update({'font.size': 20})
-
-        fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(32,6))
-
-        if np.any(np.iscomplex(z)):
-            ax[0].plot(np.abs(z), color='gray')
-        else:
-            ax[0].plot(z, color = 'gray')
-
-        ax[0].set_title('recovered latent representation (z)')
-
-        ax[1].plot(xr, color = 'gray')
-        ax[1].set_title('reconstructed signal (xr)')
-
-        fig.tight_layout()
-        plt.show()
-
-        matplotlib.rcParams.update({'font.size': 12})
+    if display:
+        plot_reconstructed(z, xr)
 
     return z, xr
 
@@ -265,7 +261,7 @@ def GridSearch_Sensing_n_Recovery(x, PSIs, ks = [0.1, 0.2, 0.5, 1.001], solver =
     COLS = 2 + len(ks)
     RMSES = []
 
-    for idx, key in enumerate(PSIs):     
+    for idx, key in enumerate(PSIs):
 
         PSI = PSIs[key]
         
@@ -282,13 +278,13 @@ def GridSearch_Sensing_n_Recovery(x, PSIs, ks = [0.1, 0.2, 0.5, 1.001], solver =
         
         psi_name = key
         
-        PSI_H = PSI.conj().T
-        PSI_INV = np.linalg.pinv(PSI)    
-        assert np.allclose(PSI_H, PSI_INV) # Unitary Matrix
-        
+        # PSI_H = PSI.conj().T
+        # PSI_INV = np.linalg.pinv(PSI)    
+        # assert np.allclose(PSI_H, PSI_INV) # Unitary Matrix        
         #print(PSI_H)
         #print(PSI_INV)
         # theoretically, PSI_H == PSI_INV
+        
         z = PSI @ xe
         # print((xx @ PSI_H).shape)
         # print("z =", z[:10])
@@ -398,7 +394,7 @@ def GridSearch_Sensing_n_Recovery(x, PSIs, ks = [0.1, 0.2, 0.5, 1.001], solver =
     return RMSES
 
 
-def Dataset_Sensing_n_Recovery (X, y = None, k = 0.2, t = 'DCT', solver = 'LASSO', fast_lasso = False, display = 'all'):
+def Dataset_Sensing_n_Recovery (X, y = None, k = 0.2, t = 'DCT', solver = 'LASSO', L1 = 0.005, display = 'all'):
     '''
     Provide a complete pipeline for an entire dataset. 
     PCA 2D visualization and multivariate KLD are evaluated.
@@ -609,4 +605,89 @@ def Dataset_Sensing_n_Recovery (X, y = None, k = 0.2, t = 'DCT', solver = 'LASSO
         ukld = Univariate_KLD( (Xc1.mean(), Xc1.std()),  (Xc2.mean(), Xc2.std()) )
         print('KLD between the orignal and reconstructed signals ( y = ' + str(c) +  '): ', round(ukld,3))
     
-    return Z, Xr
\ No newline at end of file
+    return Z, Xr
+
+
+def plot_reconstructed(z, xr):
+    '''
+    Plot the reconstructed latent representation z and the reconstructed signal xr
+    '''
+    matplotlib.rcParams.update({'font.size': 20})
+
+    fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(32,6))
+
+    if np.any(np.iscomplex(z)):
+        ax[0].plot(np.abs(z), color='gray')
+    else:
+        ax[0].plot(z, color = 'gray')
+    ax[0].set_title('recovered latent representation (z)')
+
+    ax[1].plot(xr, color = 'gray')
+    ax[1].set_title('reconstructed signal (xr)')
+
+    fig.tight_layout()
+    plt.show()
+
+    matplotlib.rcParams.update({'font.size': 12})
+
+def GridSearch_Dataset_Sensing_n_Recovery(X, y, 
+
+ts = ['IDM', 'DCT', 'DFT'], 
+ks = [0.01, 0.05, 0.1, 0.3], 
+solvers = ['LASSO', 'OMP', 'BP'],
+L1s = [0.1, 1, 10]):
+    '''
+    Grid search for best hparams for non-adaptive CS, using classification accuracy.
+
+    Parameters
+    ----------
+    ts : transform bases, e.g., DCT, DFT
+    solvers : LASSO, OMP, etc.
+    '''
+
+    if X.shape[1] % 2 == 1:
+        X = X[:,:-1] # make columns even. DCT requires even input.
+
+    scaler = StandardScaler()  # MinMaxScaler() # StandardScaler()
+    X = scaler.fit_transform(X)
+
+    clf = LogisticRegressionCV(cv=5).fit(X, y)
+    print('LogisticRegressionCV score on entire dataset:', clf.score(X, y))
+
+    dic = {}
+    best_acc = 0
+    best_hparams = ()
+
+    for k in ks:
+        for t in ts:
+            for solver in solvers:                
+                for L1 in (L1s if solver == 'LASSO' else [0]): # L1 is only used for LASSO solver
+
+                    hparam = (k, t, solver, L1)
+                    print('\nhyper-parameters: ', str(hparam))
+                    Xr = []
+                    Z = []
+                
+                    try:
+                        for i in tqdm(range(len(X))):
+
+                            x = X[i] # X[i,:].ravel().tolist()[0] # get the i-th sample
+
+                            z, xr = Sensing_n_Recovery(x, k = k, t=t, solver = solver, L1 = L1, display = False) # display = (i==0)
+                            Xr.append(xr)
+                            Z.append(z)
+
+                        acc = clf.score(np.array(Xr), y)
+                        dic[str(hparam)] = acc
+                        print("ACC = ", acc)
+
+                        if best_acc < acc:
+                            print('Update best hparams. Acc improved from {} to {}'.format(best_acc, acc))
+                            best_acc = acc
+                            plot_reconstructed(z, scaler.inverse_transform(xr.reshape(1,-1))[0] )
+                            best_hparams = hparam
+
+                    except Exception as e:
+                        print(str(hparam), e)
+
+    return dic, best_hparams, best_acc
\ No newline at end of file
